from __future__ import annotations

import hashlib
import subprocess
import sys
from dataclasses import dataclass
from functools import wraps
from pathlib import Path
from typing import Any, Callable, TypeVar
from urllib.parse import urlparse

import openapi_fal_rest.api.files.file_exists as file_exists_api
import openapi_fal_rest.models.file_spec as file_spec_model
from fal.rest_client import REST_CLIENT
from fal.toolkit import mainify

if sys.version_info >= (3, 11):
    from typing import Concatenate
else:
    from typing_extensions import Concatenate

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

from urllib.request import Request, urlopen

import fal.flags as flags
from fal.api import FalServerlessError, InternalFalServerlessError, function

# TODO: how can we randomize the user agent to avoid being blocked?
TEMP_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:21.0) Gecko/20100101 Firefox/21.0"
}

ArgsT = ParamSpec("ArgsT")
ReturnT = TypeVar("ReturnT")


class DownloadError(Exception):
    pass


def file_exists(
    file_path: Path | str, calculate_checksum: bool
) -> file_spec_model.FileSpec | None:
    file_str = str(file_path)
    response = file_exists_api.sync_detailed(
        file_str,
        client=REST_CLIENT,
        calculate_checksum=calculate_checksum,
    )

    if response.status_code == 200:
        return response.parsed  # type: ignore
    elif response.status_code == 404:
        return None
    else:
        raise InternalFalServerlessError(
            f"Failed to check if file exists: {response.status_code} {response.content.decode()}"
        )


@mainify
@dataclass
class FileSpec:
    path: Path
    size: int
    is_file: bool
    checksum_sha256: str | None = None
    checksum_md5: str | None = None

    @classmethod
    def from_model(cls, file_spec: file_spec_model.FileSpec) -> FileSpec:
        return cls(
            path=Path(file_spec.path),
            size=file_spec.size,
            is_file=file_spec.is_file,
            checksum_sha256=file_spec.checksum_sha256
            if file_spec.checksum_sha256
            else None,
            checksum_md5=file_spec.checksum_md5 if file_spec.checksum_md5 else None,
        )


def setup(
    file_path: Path | str,
    checksum_sha256: str | None = None,
    checksum_md5: str | None = None,
    force: bool = False,
    *,
    _func_name: str | None = None,
    **isolated_config: Any,
):
    file_path = Path(file_path)

    def wrapper(
        func: Callable[Concatenate[ArgsT], ReturnT]
    ) -> Callable[Concatenate[ArgsT], Path]:
        name = _func_name or func.__name__

        @wraps(func)
        def internal_wrapper(
            *args: ArgsT.args,
            **kwargs: ArgsT.kwargs,
        ) -> Path:
            checksum = bool(checksum_sha256 or checksum_md5)

            file = file_exists(file_path, calculate_checksum=checksum)

            if not file or force or flags.FORCE_SETUP:
                config = {
                    "machine_type": "S",
                    "keep_alive": 10,
                    **isolated_config,
                }
                function(**config)(func)(*args, **kwargs)  # type: ignore

                file = file_exists(file_path, calculate_checksum=checksum)

            if not file:
                raise FalServerlessError(
                    f"Setup function {name} did not create expected path."
                )

            if checksum:
                if not file.is_file:
                    raise FalServerlessError(
                        f"Setup function {name} created a directory instead of a file."
                    )

                if checksum_sha256 and file.checksum_sha256 != checksum_sha256:
                    raise FalServerlessError(
                        f"Setup function {name} created file with unexpected SHA256 checksum.\n"
                        f"Expected {checksum_sha256} but got {file.checksum_sha256}"
                    )

                if checksum_md5 and file.checksum_md5 != checksum_md5:
                    raise FalServerlessError(
                        f"Setup function {name} created file with unexpected MD5 checksum.\n"
                        f"Expected {checksum_md5} but got {file.checksum_md5}"
                    )

            return Path(file.path)

        return internal_wrapper

    return wrapper


def _get_file_name_from_url(url: str):
    # TODO: Reduce number of requests to one for downloading a file
    request = Request(url, headers=TEMP_HEADERS)
    response = urlopen(request)

    file_name = response.headers.get_filename()

    file_name = str(file_name or "")

    if not file_name:
        url_path = urlparse(url).path
        file_name = Path(url_path).stem

    # URL without path
    if not file_name:
        file_name = hashlib.sha256(url.encode("utf-8")).hexdigest()

    return file_name


def download_file(
    url: str,
    target_dir: str | Path,
    *,
    file_name: str | None = None,
    checksum_sha256: str | None = None,
    checksum_md5: str | None = None,
    force: bool = False,
    _func_name: str = "download_file",
):
    """
    Download a file from a given URL using specified download tool.

    Args:
        url: The URL of the file to be downloaded.
        target_dir: The directory to save the downloaded file, either as a
            string path or a Path object.
        file_name: The name to be used for the downloaded file. If None, the
            name will be retrieved from the server. Defaults to None.
        checksum_sha256: SHA-256 checksum value to verify the downloaded file's
            integrity. Defaults to None.
        checksum_md5: MD5 checksum value to verify the downloaded file's
            integrity. Defaults to None.
        force: If True, force re-download even if the file already exists.
            Defaults to False.
        _func_name: Name of the function to use for debugging purposes.
            Defaults to 'download_file'.

    Raises:
        ValueError: If an unsupported download tool is specified.
        Exception: If any error occurs during the download process,
            including checksum validation failure.

    Returns:
        The path where the downloaded file has been saved.
    """
    file_name = file_name or _get_file_name_from_url(url)
    if "/" in file_name:
        raise ValueError("file name can not contain a slash.")

    target_path = Path(target_dir) / file_name

    @setup(
        target_path,
        checksum_sha256,
        checksum_md5,
        force=force,
        # isolated configs
        requirements=["urllib3"],
        _func_name=_func_name,
    )
    def download():
        print(f"Downloading {url} to {target_path}")

        # Make sure the directory exists
        target_path.parent.mkdir(parents=True, exist_ok=True)

        try:
            _download_file_python(url=url, target_path=target_path)
        except Exception as e:
            msg = f"Failed to download {url} to {target_path}"

            target_path.unlink(missing_ok=True)

            raise DownloadError(msg) from e

    return download()


def _download_file_python(url: str, target_path: Path) -> Path:
    """Download a file from a given URL and save it to a specified path using a
    Python interface.

    Args:
        url: The URL of the file to be downloaded.
        target_path: The path where the downloaded file will be saved.

    Returns:
        The path where the downloaded file has been saved.
    """
    import shutil
    import tempfile

    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        try:
            file_path = temp_file.name

            for (progress, total_size) in _stream_url_data_to_file(url, temp_file.name):
                if total_size:
                    progress_msg = f"Downloading {url}... {progress:.2%}"
                else:
                    progress_msg = f"Downloading {url}... {progress:.2f} MB"
                print(progress_msg, end="\r\n")

            # Move the file when the file is downloaded completely. Since the
            # file used is temporary, in a case of an interruption, the downloaded
            # content will be lost. So, it is safe to redownload the file in such cases.
            shutil.move(file_path, target_path)

        except:
            Path(temp_file.name).unlink(missing_ok=True)
            raise

    return target_path


def _stream_url_data_to_file(url: str, file_path: str, chunk_size_in_mb: int = 64):
    """Download data from a URL and stream it to a file.

    Note:
        - This function sets a User-Agent header to mimic a web browser to
            prevent issues with some websites.
        - It downloads the file in chunks to save memory and ensures the file
            is only moved when the download is complete.

    Args:
        request: The Request object representing the URL to download from.
        file_path: The path to the file where the downloaded data will be saved.
        chunk_size_in_mb: The size of each download chunk in megabytes.
            Defaults to 64.

    Yields:
        A tuple containing two elements:
        - float: The progress of the download as a percentage (0.0 to 1.0) if
            the total size is known. Else, equals to the downloaded size in MB.
        - int: The total size of the downloaded content in bytes. If the total
            size is not known (e.g., the server doesn't provide a
            'content-length' header), the second element is 0.
    """
    ONE_MB = 1024**2

    request = Request(url, headers=TEMP_HEADERS)

    received_size = 0
    total_size = 0

    with urlopen(request) as response, open(file_path, "wb") as f_stream:
        total_size = int(response.headers.get("content-length", total_size))
        while data := response.read(chunk_size_in_mb * ONE_MB):
            f_stream.write(data)

            received_size = f_stream.tell()

            if total_size:
                progress = received_size / total_size
            else:
                progress = received_size / ONE_MB
            yield progress, total_size

    # Check if received size matches the expected total size
    if total_size and received_size < total_size:
        raise DownloadError("Received less data than expected from the server.")


def download_weights(
    url: str,
    *,
    checksum_sha256: str | None = None,
    checksum_md5: str | None = None,
    force: bool = False,
):
    """Download pre-trained weights from a given URL and save them to a
    specified location.

    This function is a convenient wrapper around the 'download_file' function,
    specialized for downloading pre-trained model weights.

    Args:
        url: The URL of the pre-trained weights file to download.
        checksum_sha256: SHA-256 checksum value to verify the downloaded file's
            integrity. Defaults to None.
        checksum_md5: MD5 checksum value to verify the downloaded file's
            integrity. Defaults to None.
        force: If True, force re-download even if the file already exists.
            Defaults to False.

    Returns:
        The path to the downloaded model weights file.

    """
    # This is not a protected path, so the user may change stuff internally
    weights_location = Path(f"/data/.fal/weights")
    weights_name = hashlib.sha256(url.encode("utf-8")).hexdigest()

    return download_file(
        url,
        target_dir=weights_location,
        file_name=weights_name,
        checksum_sha256=checksum_sha256,
        checksum_md5=checksum_md5,
        force=force,
        _func_name="download_weights",
    )


def download_repo(
    https_url: str,
    *,
    commit_hash: str | None = None,
    target_dir: str | Path | None = None,
    repo_name: str | None = None,
    checksum_sha256: str | None = None,
    checksum_md5: str | None = None,
    force: bool = False,
):
    """
    Download a repository from a given HTTPS URL and optionally a specific
    commit hash.

    Args:
        https_url: The HTTPS URL of the repository to be downloaded.
        commit_hash: Optional commit hash or reference to checkout after
            download. Defaults to None, which means the latest commit will be
            used.
        target_dir: The local directory where the repository will be
            cloned or saved. This can be specified as a string path or a Path
            object. Defaults to None, which will use a predefined directory.
        repo_name: The name to be used for the cloned repository directory.
            If None, the default name of the repository will be used.
            Defaults to None.
        checksum_sha256: SHA-256 checksum value to verify the integrity of the
            downloaded repository. Defaults to None, skipping checksum
            validation.
        checksum_md5: MD5 checksum value to verify the integrity of the
            downloaded repository. Defaults to None, skipping checksum
            validation.
        force: If True, force re-download or re-clone even if the repository
            already exists in the specified local directory. Defaults to False.

    Returns:
        The path where the downloaded repository has been cloned or saved.
    """
    if not repo_name:
        repo_name = Path(https_url).stem

    if not target_dir:
        target_dir = Path("/data/.fal/repos")

    local_repo_path = Path(target_dir) / repo_name
    local_repo_path_str = str(local_repo_path)

    @setup(
        local_repo_path,
        checksum_sha256,
        checksum_md5,
        force=force,
        _func_name="download_repo",
    )
    def download():
        try:
            if local_repo_path.exists():
                print("Removing existing repository.")
                remove_command = ["rm", "-rf", local_repo_path_str]
                subprocess.run(remove_command, check=True)

            print(f"Downloading repository '{https_url}' to {local_repo_path}")
            clone_command = ["git", "clone", https_url, local_repo_path_str]
            subprocess.run(clone_command, check=True)

            if commit_hash:
                checkout_command = ["git", "checkout", commit_hash]
                subprocess.run(checkout_command, cwd=local_repo_path, check=True)

        except Exception as e:
            print(
                f"Failed to download repository '{https_url}' to '{local_repo_path}' .",
                e,
                sep="\n",
            )

            subprocess.run(["rm", "-rf", local_repo_path_str])

    output_path = download()
    return output_path



from __future__ import annotations

import json
from datetime import datetime
from http import HTTPStatus
from sys import argv
from typing import Literal
from uuid import uuid4

import click
import fal.auth as auth
import grpc
import openapi_fal_rest.api.billing.get_user_details as get_user_details
import openapi_fal_rest.api.logs.list_since as list_logs
from fal import api, sdk
from fal.console import console
from fal.exceptions import ApplicationExceptionHandler
from fal.logging import get_logger, set_debug_logging
from fal.logging.isolate import IsolateLogPrinter
from fal.logging.trace import get_tracer
from fal.rest_client import REST_CLIENT
from fal.sdk import KeyScope
from isolate.logs import Log, LogLevel, LogSource
from rich.table import Table

DEFAULT_HOST = "api.alpha.fal.ai"
HOST_ENVVAR = "FAL_HOST"

DEFAULT_PORT = "443"
PORT_ENVVAR = "FAL_PORT"

DEBUG_ENABLED = False


log = get_logger(__name__)


class ExecutionInfo:
    debug: bool
    invocation_id: str

    def __init__(self, debug=False):
        self.debug = debug
        self.invocation_id = str(uuid4())


class MainGroup(click.Group):
    """A custom implementation of the top-level group
    (i.e. called on all commands and subcommands).

    This implementation allows for centralized behavior, including
    exception handling.
    """

    _exception_handler = ApplicationExceptionHandler()

    _tracer = get_tracer(__name__)

    def invoke(self, ctx):
        execution_info = ExecutionInfo(debug=ctx.params["debug"])
        qualified_name = " ".join([ctx.info_name] + argv[1:])
        invocation_id = execution_info.invocation_id
        set_debug_logging(execution_info.debug)

        with self._tracer.start_as_current_span(
            qualified_name, attributes={"invocation_id": invocation_id}
        ):
            try:
                log.debug(
                    f"Executing command: {qualified_name}",
                    command=qualified_name,
                )
                return super().invoke(ctx)
            except Exception as exception:
                log.error(exception)
                if execution_info.debug:
                    # Here we supress detailed errors on click lines because
                    # they're mostly decorator calls, irrelevant to the dev's error tracing
                    console.print_exception(suppress=[click])
                    console.print()
                    console.print(
                        f"The [markdown.code]invocation_id[/] for this operation is: [white]{invocation_id}[/]"
                    )
                else:
                    self._exception_handler.handle(exception)

    def add_command(
        self,
        cmd: click.Command,
        name: str | None = None,
        aliases: list[str] | None = None,
    ) -> None:
        name = name or cmd.name
        assert name, "Command must have a name"

        if not aliases:
            aliases = []

        if aliases:
            # Add aliases to the help text
            aliases_str = "Alias: " + ", ".join([name, *aliases])
            cmd.help = (cmd.help or "") + "\n\nAlias: " + ", ".join([name, *aliases])
            cmd.short_help = (
                (cmd.short_help or "") + "(Alias: " + ", ".join(aliases) + ")"
            )

        super().add_command(cmd, name)
        alias_cmd = AliasCommand(cmd)

        for alias in aliases:
            self.add_command(alias_cmd, alias)


class AliasCommand(click.Command):
    def __init__(self, wrapped):
        self._wrapped = wrapped

    def __getattribute__(self, __name: str):
        if __name == "_wrapped":
            # To be able to call `self._wrapped` below
            return super().__getattribute__(__name)

        if __name == "hidden":
            return True

        return self._wrapped.__getattribute__(__name)


@click.group(cls=MainGroup)
@click.option(
    "--debug", is_flag=True, help="Enable detailed errors and verbose logging."
)
@click.version_option()
def cli(debug):
    pass


###### Auth group ######
@click.group
def auth_cli():
    pass


@auth_cli.command(name="login")
def auth_login():
    auth.login()
    try:
        client = sdk.FalServerlessClient(f"{DEFAULT_HOST}:{DEFAULT_PORT}")
        with client.connect() as connection:
            connection.list_worker_status()
    except grpc.RpcError as e:
        if "Insufficient permissions" in e.details():
            console.print(e.details())
        else:
            raise e


@auth_cli.command(name="logout")
def auth_logout():
    auth.logout()


@auth_cli.command(name="hello", hidden=True)
def auth_test():
    """
    To test auth.
    """
    print(f"Hello, {auth.USER.info['name']} - '{auth.USER.info['sub']}'")


###### Key group ######
@click.group
@click.option("--host", default=DEFAULT_HOST, envvar=HOST_ENVVAR)
@click.option("--port", default=DEFAULT_PORT, envvar=PORT_ENVVAR, hidden=True)
@click.pass_context
def key_cli(ctx, host: str, port: str):
    ctx.obj = sdk.FalServerlessClient(f"{host}:{port}")


@key_cli.command(name="generate", no_args_is_help=True)
@click.option(
    "--scope",
    default=None,
    required=True,
    type=click.Choice([KeyScope.ADMIN.value, KeyScope.API.value]),
    help="The privilage scope of the key.",
)
@click.option(
    "--alias",
    default=None,
    help="An alias for the key.",
)
@click.pass_obj
def key_generate(client: sdk.FalServerlessClient, scope: str, alias: str | None):
    with client.connect() as connection:
        parsed_scope = KeyScope(scope)
        result = connection.create_user_key(parsed_scope, alias)
        print(
            f"Generated key id and key secret, with the scope `{scope}`.\n"
            "This is the only time the secret will be visible.\n"
            "You will need to generate a new key pair if you lose access to this secret."
        )
        print(f"FAL_KEY_ID='{result[1]}'\nFAL_KEY_SECRET='{result[0]}'")


@key_cli.command(name="list")
@click.pass_obj
def key_list(client: sdk.FalServerlessClient):
    table = Table(title="Keys")
    table.add_column("Key ID")
    table.add_column("Created At")
    table.add_column("Scope")
    table.add_column("Alias")

    with client.connect() as connection:
        keys = connection.list_user_keys()
        for key in keys:
            table.add_row(
                key.key_id, str(key.created_at), str(key.scope.value), key.alias
            )

    console.print(table)


@key_cli.command(name="revoke")
@click.argument("key_id", required=True)
@click.pass_obj
def key_revoke(client: sdk.FalServerlessClient, key_id: str):
    with client.connect() as connection:
        connection.revoke_user_key(key_id)


###### Usage group ######
@click.group
@click.option("--host", default=DEFAULT_HOST, envvar=HOST_ENVVAR)
@click.option("--port", default=DEFAULT_PORT, envvar=PORT_ENVVAR, hidden=True)
@click.pass_context
def usage_cli(ctx, host: str, port: str):
    ctx.obj = sdk.FalServerlessClient(f"{host}:{port}")


@usage_cli.command(name="workers")
@click.option("--user", hidden=True, default=None)
@click.pass_obj
def usage_worker_status(client: sdk.FalServerlessClient, user: str | None):
    table = Table(title="Worker status")
    table.add_column("Worker ID")
    table.add_column("User ID")
    table.add_column("Machine type")
    table.add_column("Start time")
    table.add_column("End time")
    table.add_column("Duration")

    with client.connect() as connection:
        for ws in connection.list_worker_status(user_id=user):
            table.add_row(
                ws.worker_id,
                ws.user_id,
                ws.machine_type,
                str(ws.start_time),
                str(ws.end_time),
                str(ws.duration),
            )

    console.print(table)


##### Function group #####
@click.group
@click.option("--host", default=DEFAULT_HOST, envvar=HOST_ENVVAR)
@click.option("--port", default=DEFAULT_PORT, envvar=PORT_ENVVAR, hidden=True)
@click.pass_context
def function_cli(ctx, host: str, port: str):
    ctx.obj = api.FalServerlessHost(f"{host}:{port}")


@function_cli.command("serve")
@click.option("--alias", default=None)
@click.option(
    "--auth",
    "auth_mode",
    type=click.Choice(["public", "private", "shared"]),
    default="private",
)
@click.argument("file_path", required=True)
@click.argument("function_name", required=True)
@click.pass_obj
def register_application(
    host: api.FalServerlessHost,
    file_path: str,
    function_name: str,
    alias: str | None,
    auth_mode: Literal["public", "private", "shared"],
):
    import runpy

    user_id = _get_user_id()

    module = runpy.run_path(file_path)
    if function_name not in module:
        raise api.FalServerlessError(f"Function '{function_name}' not found in module")

    isolated_function = module[function_name]
    gateway_options = isolated_function.options.gateway
    if "serve" not in gateway_options and "exposed_port" not in gateway_options:
        raise api.FalServerlessError(
            "One of `serve` or `exposed_port` options needs to be specified in the isolated annotation to register a function"
        )
    elif (
        "exposed_port" in gateway_options
        and str(gateway_options.get("exposed_port")) != "8080"
    ):
        raise api.FalServerlessError(
            "Must expose port 8080 for now. This will be configurable in the future."
        )

    max_concurrency = gateway_options.get("max_concurrency")
    id = host.register(
        func=isolated_function.func,
        options=isolated_function.options,
        application_name=alias,
        application_auth_mode=auth_mode,
        max_concurrency=max_concurrency,
    )

    if id:
        # TODO: should we centralize this URL format?
        gateway_host = host.url.replace("api.", "gateway.")
        gateway_host = remove_http_and_port_from_url(gateway_host)

        if alias:
            console.print(
                f"Registered a new revision for function '{alias}' (revision='{id}')."
            )
            console.print(f"URL: https://{user_id}-{alias}.{gateway_host}")
        else:
            console.print(f"Registered anonymous function '{id}'.")
            console.print(f"URL: https://{user_id}-{id}.{gateway_host}")


@function_cli.command("schedule")
@click.argument("cron_string", required=True)
@click.argument("file_path", required=True)
@click.argument("function_name", required=True)
@click.pass_obj
def register_schedulded(
    host: api.FalServerlessHost, cron_string: str, file_path: str, function_name: str
):
    import runpy

    module = runpy.run_path(file_path)
    isolated_function = module[function_name]

    cron_id = host.schedule(
        func=isolated_function.func, cron=cron_string, options=isolated_function.options
    )
    if cron_id:
        console.print(cron_id)


@function_cli.command("logs")
@click.option("--lines", default=100)
@click.option("--url", default=None)
@click.pass_obj
def get_logs(
    host: api.FalServerlessHost, lines: int | None = 100, url: str | None = None
):
    log_printer = IsolateLogPrinter(debug=True)
    logs_response = list_logs.sync_detailed(
        client=REST_CLIENT, limit=lines, url_query=url
    )
    if not logs_response.status_code == 200 or type(logs_response.parsed) != list:
        raise api.FalServerlessError(str(logs_response.parsed))
    if len(logs_response.parsed) == 0:
        console.print("No logs found")
    for log in logs_response.parsed:
        app = log.app or "fal"

        log_printer.print(
            Log(
                message=f"{app}: {log.message}",
                source=LogSource.USER,
                level=LogLevel[log.level],
                timestamp=datetime.fromisoformat(log.timestamp),
            )
        )


##### Alias group #####
@click.group
@click.option("--host", default=DEFAULT_HOST, envvar=HOST_ENVVAR)
@click.option("--port", default=DEFAULT_PORT, envvar=PORT_ENVVAR, hidden=True)
@click.pass_context
def alias_cli(ctx, host: str, port: str):
    ctx.obj = api.FalServerlessClient(f"{host}:{port}")


@alias_cli.command("list")
@click.pass_obj
def alias_list(client: api.FalServerlessClient):
    with client.connect() as connection:
        table = Table(title="Function Aliases")
        table.add_column("Alias")
        table.add_column("Revision")
        table.add_column("Auth")
        table.add_column("Max Concurrency")

        for app_alias in connection.list_aliases():
            table.add_row(
                app_alias.alias,
                app_alias.revision,
                app_alias.auth_mode,
                str(app_alias.max_concurrency),
            )

    console.print(table)


@alias_cli.command("scale")
@click.argument("alias", required=True)
@click.argument("max_concurrency", required=True, type=int)
@click.pass_obj
def alias_scale(client: api.FalServerlessClient, alias: str, max_concurrency: int):
    with client.connect() as connection:
        connection.scale(application_name=alias, max_concurrency=max_concurrency)


@alias_cli.command("update")
@click.argument("alias", required=True)
@click.option("--keep-alive", type=int)
@click.pass_obj
def alias_update(client: api.FalServerlessClient, alias: str, keep_alive: int | None):
    with client.connect() as connection:
        if not keep_alive:
            console.log("No parameters for update were provided, ignoring.")
            return

        connection.update_application(application_name=alias, keep_alive=keep_alive)


##### Crons group #####
@click.group
@click.option("--host", default=DEFAULT_HOST, envvar=HOST_ENVVAR)
@click.option("--port", default=DEFAULT_PORT, envvar=PORT_ENVVAR, hidden=True)
@click.pass_context
def crons_cli(ctx, host: str, port: str):
    ctx.obj = api.FalServerlessHost(f"{host}:{port}")


@crons_cli.command(name="list")
@click.pass_obj
def list_scheduled(host: api.FalServerlessHost):
    table = Table(title="Cronjobs")
    table.add_column("Cron ID")
    table.add_column("Cron")
    table.add_column("ETA next run")
    table.add_column("State")
    for cron in host._connection.list_scheduled_runs():
        state_string = ["Not Active", "Active"][cron.active]
        table.add_row(cron.cron_id, cron.cron_string, str(cron.next_run), state_string)

    console.print(table)


@crons_cli.command(name="activations")
@click.argument("cron_id", required=True)
@click.argument("limit", default=15)
@click.pass_obj
def list_activations(host: api.FalServerlessHost, cron_id: str, limit: int = 15):
    table = Table(title="Cron activations")
    table.add_column("Activation ID")
    table.add_column("Start Date")
    table.add_column("Finish Date")

    for activation in host._connection.list_run_activations(cron_id)[:limit]:
        table.add_row(
            str(activation.activation_id),
            str(activation.started_at),
            str(activation.finished_at),
        )

    console.print(table)


@crons_cli.command(name="logs")
@click.argument("cron_id", required=True)
@click.argument("activation_id", required=True)
@click.pass_obj
def print_logs(host: api.FalServerlessHost, cron_id: str, activation_id: str):
    logs = host._connection.get_activation_logs(cron_id, activation_id)
    if not logs:
        console.print(f"No logs found for activation {activation_id}")
        return
    log_printer = IsolateLogPrinter(debug=True)
    for log in logs:
        log_printer.print(log)


@crons_cli.command("cancel")
@click.argument("cron_id", required=True)
@click.pass_obj
def cancel_scheduled(host: api.FalServerlessHost, cron_id: str):
    host._connection.cancel_scheduled_run(cron_id)
    console.print("Cancelled", repr(cron_id))


@click.group
@click.option("--host", default=DEFAULT_HOST, envvar=HOST_ENVVAR)
@click.option("--port", default=DEFAULT_PORT, envvar=PORT_ENVVAR, hidden=True)
@click.pass_context
def secrets_cli(ctx, host: str, port: str):
    ctx.obj = sdk.FalServerlessClient(f"{host}:{port}")


@secrets_cli.command("list")
@click.pass_obj
def list_secrets(client: api.FalServerlessClient):
    table = Table(title="Secrets")
    table.add_column("Secret Name")
    table.add_column("Created At")

    with client.connect() as connection:
        for secret in connection.list_secrets():
            table.add_row(secret.name, str(secret.created_at))

    console.print(table)


@secrets_cli.command("set")
@click.argument("secret_name", required=True)
@click.argument("secret_value", required=True)
@click.pass_obj
def set_secret(client: api.FalServerlessClient, secret_name: str, secret_value: str):
    with client.connect() as connection:
        connection.set_secret(secret_name, secret_value)
        console.print(f"Secret '{secret_name}' has set")


@secrets_cli.command("delete")
@click.argument("secret_name", required=True)
@click.pass_obj
def delete_secret(client: api.FalServerlessClient, secret_name: str):
    with client.connect() as connection:
        connection.delete_secret(secret_name)
        console.print(f"Secret '{secret_name}' has deleted")


cli.add_command(auth_cli, name="auth")
cli.add_command(key_cli, name="key", aliases=["keys"])
cli.add_command(function_cli, name="function", aliases=["fn"])
cli.add_command(alias_cli, name="alias", aliases=["aliases"])
cli.add_command(crons_cli, name="cron", aliases=["crons"])
cli.add_command(usage_cli, name="usage")
cli.add_command(secrets_cli, name="secret", aliases=["secrets"])


def remove_http_and_port_from_url(url):
    # Remove http://
    if "http://" in url:
        url = url.replace("http://", "")

    # Remove https://
    if "https://" in url:
        url = url.replace("https://", "")

    # Remove port information
    url_parts = url.split(":")
    if len(url_parts) > 1:
        url = url_parts[0]

    return url


# dbt-fal commands to be errored out
DBT_FAL_COMMAND_NOTICE = """
The dbt tool `fal` and `dbt-fal` adapter have been merged into a single tool.
Please use the new `dbt-fal` command line tool instead.
Running `pip install dbt-fal` will install the new tool and the adapter alongside.
Then run your command like

    dbt-fal <command>

"""


@cli.command("run", context_settings={"ignore_unknown_options": True})
@click.argument("any", nargs=-1, type=click.UNPROCESSED)
def dbt_run(any):
    raise click.BadArgumentUsage(DBT_FAL_COMMAND_NOTICE)


@cli.command("flow", context_settings={"ignore_unknown_options": True})
@click.argument("any", nargs=-1, type=click.UNPROCESSED)
def dbt_flow(any):
    raise click.BadArgumentUsage(DBT_FAL_COMMAND_NOTICE)


def _get_user_id() -> str:
    try:
        user_details_response = get_user_details.sync_detailed(
            client=REST_CLIENT,
        )
    except Exception as e:
        raise api.FalServerlessError(f"Error fetching user details: {str(e)}")

    if user_details_response.status_code != HTTPStatus.OK:
        try:
            content = json.loads(user_details_response.content.decode("utf8"))
        except:
            raise api.FalServerlessError(
                f"Error fetching user details: {user_details_response}"
            )
        else:
            raise api.FalServerlessError(content["detail"])
    try:
        full_user_id = user_details_response.parsed.user_id
        user_id = full_user_id.split("|")[1]
        return user_id
    except Exception as e:
        raise api.FalServerlessError(f"Could not parse the user data: {e}")


if __name__ == "__main__":
    cli()



from __future__ import annotations

from dataclasses import dataclass
from mimetypes import guess_extension, guess_type
from typing import Literal
from uuid import uuid4

from fal.toolkit.mainify import mainify


@mainify
class FileData:
    data: bytes
    content_type: str
    file_name: str

    def __init__(
        self, data: bytes, content_type: str | None = None, file_name: str | None = None
    ):
        self.data = data
        if content_type is None and file_name is not None:
            content_type, _ = guess_type(file_name or "")

        # Ultimately fallback to a generic binary file mime type
        self.content_type = content_type or "application/octet-stream"

        if file_name is None:
            extension = guess_extension(self.content_type, strict=False) or ".bin"
            self.file_name = f"{uuid4().hex}{extension}"
        else:
            self.file_name = file_name


RepositoryId = Literal["fal", "in_memory", "gcp_storage"]


@mainify
@dataclass
class FileRepository:
    def save(self, data: FileData) -> str:
        raise NotImplementedError()



from __future__ import annotations

from functools import wraps

import dill
from dill import _dill
from fal.toolkit import mainify

# each @fal.function gets added to this set so that we can
# mainify the module this function is in
_MODULES: set[str] = set()
_PACKAGES: set[str] = set()


@mainify
def _pydantic_make_field(kwargs):
    from pydantic.fields import ModelField

    return ModelField(**kwargs)


@mainify
def _pydantic_make_private_field(kwargs):
    from pydantic.fields import ModelPrivateAttr

    return ModelPrivateAttr(**kwargs)


# this allows us to record all the "isolated" function and then mainify everything in
# module they exist
@wraps(_dill._locate_function)
def by_value_locator(obj, pickler=None, og_locator=_dill._locate_function):
    module_name = getattr(obj, "__module__", None)
    if module_name is not None:
        # If it is coming from the same module, directly allow
        # it to be pickled.
        if module_name in _MODULES:
            return False

        package_name, *_ = module_name.partition(".")
        # If it is coming from the same package, then do the same.
        if package_name in _PACKAGES:
            return False

    og_result = og_locator(obj, pickler)
    return og_result


_dill._locate_function = by_value_locator


def add_serialization_listeners_for(obj):
    module_name = getattr(obj, "__module__", None)
    if not module_name:
        return None

    _MODULES.add(module_name)

    if "." in module_name:
        package_name, *_ = module_name.partition(".")
        _PACKAGES.add(package_name)


@mainify
def patch_pydantic_field_serialization():
    # Cythonized pydantic fields can't be serialized automatically, so we are
    # have a special case handling for them that unpacks it to a dictionary
    # and then reloads it on the other side.
    import dill

    try:
        import pydantic.fields
    except ImportError:
        return

    @dill.register(pydantic.fields.ModelField)
    def _pickle_model_field(
        pickler: dill.Pickler,
        field: pydantic.fields.ModelField,
    ) -> None:
        args = {
            "name": field.name,
            # outer_type_ is the original type for ModelFields,
            # while type_ can be updated later with the nested type
            # like int for List[int].
            "type_": field.outer_type_,
            "class_validators": field.class_validators,
            "model_config": field.model_config,
            "default": field.default,
            "default_factory": field.default_factory,
            "required": field.required,
            "alias": field.alias,
            "field_info": field.field_info,
        }
        pickler.save_reduce(_pydantic_make_field, (args,), obj=field)

    @dill.register(pydantic.fields.ModelPrivateAttr)
    def _pickle_model_private_attr(
        pickler: dill.Pickler,
        field: pydantic.fields.ModelPrivateAttr,
    ) -> None:
        args = {
            "default": field.default,
            "default_factory": field.default_factory,
        }
        pickler.save_reduce(_pydantic_make_private_field, (args,), obj=field)


@mainify
def patch_pydantic_class_attributes():
    # Dill attempts to modify the __class__ of deserialized pydantic objects
    # on this side but it meets with a rejection from pydantic's semantics since
    # __class__ is not recognized as a proper dunder attribute.
    try:
        import pydantic.utils
    except ImportError:
        return

    pydantic.utils.DUNDER_ATTRIBUTES.add("__class__")


@mainify
def patch_dill():
    import dill

    dill.settings["recurse"] = True
    patch_pydantic_class_attributes()
    patch_pydantic_field_serialization()



from __future__ import annotations

import enum
import os
from contextlib import ExitStack
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Callable, Generic, Iterator, Literal, TypeVar

import grpc
import isolate_proto
from fal import flags
from fal._serialization import patch_dill
from fal.auth import USER
from fal.logging import get_logger
from fal.logging.trace import TraceContextInterceptor
from isolate.connections.common import is_agent
from isolate.logs import Log
from isolate.server.interface import from_grpc, to_serialized_object
from isolate_proto.configuration import GRPC_OPTIONS

ResultT = TypeVar("ResultT")
InputT = TypeVar("InputT")
UNSET = object()

_DEFAULT_SERIALIZATION_METHOD = "dill"
FAL_SERVERLESS_DEFAULT_KEEP_ALIVE = 10

log = get_logger(__name__)

patch_dill()


class ServerCredentials:
    def to_grpc(self) -> grpc.ChannelCredentials:
        raise NotImplementedError

    @property
    def extra_options(self) -> list[tuple[str, str]]:
        return GRPC_OPTIONS


class LocalCredentials(ServerCredentials):
    def to_grpc(self) -> grpc.ChannelCredentials:
        return grpc.local_channel_credentials()


class RemoteCredentials(ServerCredentials):
    def to_grpc(self) -> grpc.ChannelCredentials:
        return grpc.ssl_channel_credentials()


@dataclass
class _GRPCMetadata(grpc.AuthMetadataPlugin):
    """Key value metadata bundle for gRPC credentials"""

    _key: str
    _value: str

    def __call__(
        self,
        context: grpc.AuthMetadataContext,
        callback: grpc.AuthMetadataPluginCallback,
    ) -> None:
        callback(((self._key, self._value),), None)


def get_default_server_credentials() -> ServerCredentials:
    if flags.TEST_MODE:
        return LocalCredentials()
    else:
        return RemoteCredentials()


class Credentials:
    # Cannot use `field` because child classes don't have default for all properties.
    server_credentials: ServerCredentials = get_default_server_credentials()

    def to_grpc(self) -> grpc.ChannelCredentials:
        return self.server_credentials.to_grpc()

    def to_headers(self) -> dict[str, str]:
        return {}


@dataclass
class FalServerlessKeyCredentials(Credentials):
    key_id: str
    key_secret: str

    def to_grpc(self) -> grpc.ChannelCredentials:
        return grpc.composite_channel_credentials(
            self.server_credentials.to_grpc(),
            grpc.metadata_call_credentials(_GRPCMetadata("auth-key", self.key_secret)),
            grpc.metadata_call_credentials(_GRPCMetadata("auth-key-id", self.key_id)),
        )

    def to_headers(self) -> dict[str, str]:
        return {"Authorization": f"Key {self.key_id}:{self.key_secret}"}


@dataclass
class AuthenticatedCredentials(Credentials):
    user = USER

    def to_grpc(self) -> grpc.ChannelCredentials:
        return grpc.composite_channel_credentials(
            self.server_credentials.to_grpc(),
            grpc.access_token_call_credentials(USER.access_token),
        )

    def to_headers(self) -> dict[str, str]:
        token = USER.bearer_token
        return {"Authorization": token}


@dataclass
class ServerlessSecret:
    name: str
    created_at: datetime


def key_credentials() -> FalServerlessKeyCredentials | None:
    # Ignore key credentials when the user forces auth by user.
    if os.environ.get("FAL_FORCE_AUTH_BY_USER") == "1":
        return None

    if "FAL_KEY_ID" in os.environ and "FAL_KEY_SECRET" in os.environ:
        return FalServerlessKeyCredentials(
            os.environ["FAL_KEY_ID"],
            os.environ["FAL_KEY_SECRET"],
        )
    else:
        return None


def _get_agent_credentials(original_credentials: Credentials) -> Credentials:
    """If running inside a fal Serverless box, use the preconfigured credentials
    instead of the user provided ones."""

    key_creds = key_credentials()
    if is_agent() and key_creds:
        return key_creds
    else:
        return original_credentials


def get_default_credentials() -> Credentials:
    if flags.AUTH_DISABLED:
        return Credentials()

    key_creds = key_credentials()
    if key_creds:
        log.debug("Using key credentials")
        return key_creds
    else:
        return AuthenticatedCredentials()


@dataclass
class FalServerlessClient:
    hostname: str
    credentials: Credentials = field(default_factory=get_default_credentials)

    def connect(self) -> FalServerlessConnection:
        return FalServerlessConnection(self.hostname, self.credentials)


class HostedRunState(Enum):
    IN_PROGRESS = 0
    SUCCESS = 1
    INTERNAL_FAILURE = 2


@dataclass
class HostedRunStatus:
    state: HostedRunState


@dataclass
class AliasInfo:
    alias: str
    revision: str
    auth_mode: str
    max_concurrency: int


@dataclass(frozen=True)
class Cron:
    cron_id: str
    cron_string: str
    next_run: datetime
    active: bool


@dataclass(frozen=True)
class ScheduledRunActivation:
    cron_id: str
    activation_id: str
    started_at: datetime
    finished_at: datetime


@dataclass
class HostedRunResult(Generic[ResultT]):
    run_id: str
    status: HostedRunStatus
    logs: list[Log] = field(default_factory=list)
    result: ResultT | None = None


@dataclass
class RegisterApplicationResult:
    result: RegisterApplicationResultType | None
    logs: list[Log] = field(default_factory=list)


@dataclass(frozen=True)
class RegisterCronResultType:
    cron_id: str


@dataclass(frozen=True)
class RegisterCronResult:
    result: RegisterCronResultType


@dataclass
class RegisterApplicationResultType:
    application_id: str


@dataclass
class UserKeyInfo:
    key_id: str
    created_at: datetime
    scope: KeyScope
    alias: str


@dataclass
class WorkerStatus:
    worker_id: str
    start_time: datetime
    end_time: datetime
    duration: timedelta
    user_id: str
    machine_type: str


class KeyScope(enum.Enum):
    ADMIN = "ADMIN"
    API = "API"

    @staticmethod
    def from_proto(
        proto: isolate_proto.CreateUserKeyRequest.Scope.ValueType | None,
    ) -> KeyScope:
        if proto is None:
            return KeyScope.API

        if proto is isolate_proto.CreateUserKeyRequest.Scope.ADMIN:
            return KeyScope.ADMIN
        elif proto is isolate_proto.CreateUserKeyRequest.Scope.API:
            return KeyScope.API
        else:
            raise ValueError(f"Unknown KeyScope: {proto}")


@from_grpc.register(isolate_proto.RegisterCronResult)
def _from_grpc_register_cron_result(
    message: isolate_proto.RegisterCronResult,
) -> RegisterCronResult:
    return RegisterCronResult(
        result=RegisterCronResultType(message.result.cron_id),
    )


@from_grpc.register(isolate_proto.CronResultType)
def _from_grpc_cron_result_type(
    message: isolate_proto.CronResultType,
) -> Cron:
    return Cron(
        cron_id=message.cron_id,
        cron_string=message.cron_string,
        next_run=message.next_run.ToDatetime(),
        active=message.is_active,
    )


@from_grpc.register(isolate_proto.AliasInfo)
def _from_grpc_alias_info(message: isolate_proto.AliasInfo) -> AliasInfo:
    if message.auth_mode is isolate_proto.ApplicationAuthMode.PUBLIC:
        auth_mode = "public"
    elif message.auth_mode is isolate_proto.ApplicationAuthMode.PRIVATE:
        auth_mode = "private"
    elif message.auth_mode is isolate_proto.ApplicationAuthMode.SHARED:
        auth_mode = "shared"
    else:
        raise ValueError(f"Unknown auth mode: {message.auth_mode}")

    return AliasInfo(
        alias=message.alias,
        revision=message.revision,
        auth_mode=auth_mode,
        max_concurrency=message.max_concurrency,
    )


@from_grpc.register(isolate_proto.RegisterApplicationResult)
def _from_grpc_register_application_result(
    message: isolate_proto.RegisterApplicationResult,
) -> RegisterApplicationResult:
    return RegisterApplicationResult(
        logs=[from_grpc(log) for log in message.logs],
        result=None
        if not message.HasField("result")
        else RegisterApplicationResultType(message.result.application_id),
    )


@from_grpc.register(isolate_proto.HostedRunStatus)
def _from_grpc_hosted_run_status(
    message: isolate_proto.HostedRunStatus,
) -> HostedRunStatus:
    return HostedRunStatus(HostedRunState(message.state))


@from_grpc.register(isolate_proto.HostedRunResult)
def _from_grpc_hosted_run_result(
    message: isolate_proto.HostedRunResult,
) -> HostedRunResult[Any]:
    if message.return_value.definition:
        return_value = from_grpc(message.return_value)
    else:
        return_value = UNSET

    return HostedRunResult(
        message.run_id,
        from_grpc(message.status),
        logs=[from_grpc(log) for log in message.logs],
        result=return_value,
    )


def _get_cron_id(run: Cron | str) -> str:
    if isinstance(run, Cron):
        return run.cron_id
    else:
        return run


@dataclass
class MachineRequirements:
    machine_type: str
    keep_alive: int = FAL_SERVERLESS_DEFAULT_KEEP_ALIVE
    base_image: str | None = None
    exposed_port: int | None = None
    scheduler: str | None = None


@dataclass
class FalServerlessConnection:
    hostname: str
    credentials: Credentials

    _stack: ExitStack = field(default_factory=ExitStack)
    _stub: isolate_proto.IsolateControllerStub | None = None

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self._stack.close()

    def close(self):
        self._stack.close()

    @property
    def stub(self) -> isolate_proto.IsolateControllerStub:
        if self._stub:
            return self._stub

        options = self.credentials.server_credentials.extra_options
        channel_creds = self.credentials.to_grpc()
        channel = self._stack.enter_context(
            grpc.secure_channel(self.hostname, channel_creds, options)
        )
        channel = grpc.intercept_channel(channel, TraceContextInterceptor())
        self._stub = isolate_proto.IsolateControllerStub(channel)
        return self._stub

    def create_user_key(self, scope: KeyScope, alias: str | None) -> tuple[str, str]:
        scope_proto = (
            isolate_proto.CreateUserKeyRequest.Scope.ADMIN
            if scope is KeyScope.ADMIN
            else isolate_proto.CreateUserKeyRequest.Scope.API
        )

        request = isolate_proto.CreateUserKeyRequest(scope=scope_proto, alias=alias)
        response = self.stub.CreateUserKey(request)
        return response.key_secret, response.key_id

    def list_user_keys(self) -> list[UserKeyInfo]:
        request = isolate_proto.ListUserKeysRequest()
        response: isolate_proto.ListUserKeysResponse = self.stub.ListUserKeys(request)
        return [
            UserKeyInfo(
                key.key_id,
                isolate_proto.datetime_from_timestamp(key.created_at),
                KeyScope.from_proto(key.scope),
                key.alias,
            )
            for key in response.user_keys
        ]

    def revoke_user_key(self, key_id) -> None:
        request = isolate_proto.RevokeUserKeyRequest(key_id=key_id)
        self.stub.RevokeUserKey(request)

    # TODO: get rid of this in favor of define_environment
    def create_environment(
        self,
        kind: str,
        configuration_options: dict[str, Any],
    ) -> isolate_proto.EnvironmentDefinition:
        assert isinstance(
            configuration_options, dict
        ), "configuration_options must be a dict"
        struct = isolate_proto.Struct()
        struct.update(configuration_options)

        return isolate_proto.EnvironmentDefinition(
            kind=kind,
            configuration=struct,
        )

    def define_environment(
        self, kind: str, **options: Any
    ) -> isolate_proto.EnvironmentDefinition:
        return self.create_environment(
            kind=kind,
            configuration_options=options,
        )

    def register(
        self,
        function: Callable[..., ResultT],
        environments: list[isolate_proto.EnvironmentDefinition],
        application_name: str | None = None,
        application_auth_mode: Literal["public", "private", "shared"] | None = None,
        *,
        max_concurrency: int | None = None,
        serialization_method: str = _DEFAULT_SERIALIZATION_METHOD,
        machine_requirements: MachineRequirements | None = None,
    ) -> Iterator[isolate_proto.RegisterApplicationResult]:
        wrapped_function = to_serialized_object(function, serialization_method)
        if machine_requirements:
            wrapped_requirements = isolate_proto.MachineRequirements(
                machine_type=machine_requirements.machine_type,
                keep_alive=machine_requirements.keep_alive,
                base_image=machine_requirements.base_image,
                exposed_port=machine_requirements.exposed_port,
                scheduler=machine_requirements.scheduler,
            )
        else:
            wrapped_requirements = None

        if application_auth_mode == "public":
            auth_mode = isolate_proto.ApplicationAuthMode.PUBLIC
        elif application_auth_mode == "shared":
            auth_mode = isolate_proto.ApplicationAuthMode.SHARED
        else:
            auth_mode = isolate_proto.ApplicationAuthMode.PRIVATE

        request = isolate_proto.RegisterApplicationRequest(
            function=wrapped_function,
            environments=environments,
            machine_requirements=wrapped_requirements,
            max_concurrency=max_concurrency,
            application_name=application_name,
            auth_mode=auth_mode,
        )
        for partial_result in self.stub.RegisterApplication(request):
            yield from_grpc(partial_result)

    def scale(self, application_name: str, max_concurrency: int | None = None) -> None:
        request = isolate_proto.ScaleApplicationRequest(
            application_name=application_name,
            max_concurrency=max_concurrency,
        )
        self.stub.ScaleApplication(request)

    def update_application(
        self, application_name: str, keep_alive: int | None = None
    ) -> None:
        request = isolate_proto.UpdateApplicationRequest(
            application_name=application_name,
            keep_alive=keep_alive,
        )
        self.stub.UpdateApplication(request)

    def run(
        self,
        function: Callable[..., ResultT],
        environments: list[isolate_proto.EnvironmentDefinition],
        *,
        serialization_method: str = _DEFAULT_SERIALIZATION_METHOD,
        machine_requirements: MachineRequirements | None = None,
        setup_function: Callable[[], InputT] | None = None,
    ) -> Iterator[HostedRunResult[ResultT]]:
        wrapped_function = to_serialized_object(function, serialization_method)
        if machine_requirements:
            wrapped_requirements = isolate_proto.MachineRequirements(
                machine_type=machine_requirements.machine_type,
                keep_alive=machine_requirements.keep_alive,
                base_image=machine_requirements.base_image,
                exposed_port=machine_requirements.exposed_port,
                scheduler=machine_requirements.scheduler,
            )
        else:
            wrapped_requirements = None

        request = isolate_proto.HostedRun(
            function=wrapped_function,
            environments=environments,
            machine_requirements=wrapped_requirements,
        )
        if setup_function:
            request.setup_func.MergeFrom(
                to_serialized_object(setup_function, serialization_method)
            )
        for partial_result in self.stub.Run(request):
            yield from_grpc(partial_result)

    def schedule_cronjob(
        self,
        application_id: str,
        cron: str,
    ) -> str:
        request = isolate_proto.RegisterCronRequest(
            application_id=application_id, cron=cron
        )
        response: isolate_proto.RegisterCronResult = self.stub.RegisterCron(request)
        return response.result.cron_id

    def list_aliases(self) -> list[AliasInfo]:
        request = isolate_proto.ListAliasesRequest()
        response: isolate_proto.ListAliasesResult = self.stub.ListAliases(request)
        return [from_grpc(alias) for alias in response.aliases]

    def list_scheduled_runs(self) -> list[Cron]:
        request = isolate_proto.ListCronsRequest()
        response: isolate_proto.ListCronsResult = self.stub.ListCrons(request)
        return [from_grpc(cron) for cron in response.crons]

    def list_run_activations(self, run: str | Cron) -> list[ScheduledRunActivation]:
        request = isolate_proto.ListActivationsRequest(cron_id=_get_cron_id(run))
        response: isolate_proto.ListActivationsResult = self.stub.ListActivations(
            request
        )
        return [
            ScheduledRunActivation(
                cron_id=_get_cron_id(run),
                activation_id=activation.activation_id,
                started_at=activation.started_at.ToDatetime(),
                finished_at=activation.finished_at.ToDatetime(),
            )
            for activation in response.activations
        ]

    def cancel_scheduled_run(self, cron_id: str) -> None:
        request = isolate_proto.CancelCronRequest(cron_id=cron_id)
        response: isolate_proto.CancelCronResult = self.stub.CancelCron(request)
        return

    def get_activation_logs(self, cron_id: str, activation_id: str) -> list[Log]:
        request = isolate_proto.GetActivationLogsRequest(
            cron_id=cron_id, activation_id=activation_id
        )
        response = self.stub.GetActivationLogs(request)
        return [from_grpc(log) for log in response.logs]

    def get_logs(
        self, lines: int | None = None, url: str | None = None
    ) -> Iterator[Log]:
        filter = isolate_proto.LogsFilter(lines=lines, url=url)
        request = isolate_proto.GetLogsRequest(filter=filter)
        for partial_result in self.stub.GetLogs(request):
            yield from_grpc(partial_result.log_entry)

    def list_worker_status(self, user_id: str | None = None) -> list[WorkerStatus]:
        request = isolate_proto.WorkerStatusListRequest(user_id=user_id)
        response = self.stub.WorkerStatusList(request)
        return [
            WorkerStatus(
                ws.worker_id,
                isolate_proto.datetime_from_timestamp(ws.start_time),
                isolate_proto.datetime_from_timestamp(ws.end_time),
                isolate_proto.timedelta_from_duration(ws.duration),
                ws.user_id,
                ws.machine_type,
            )
            for ws in response.worker_status
        ]

    def set_secret(self, name: str, value: str) -> None:
        request = isolate_proto.SetSecretRequest(name=name, value=value)
        self.stub.SetSecret(request)

    def delete_secret(self, name: str) -> None:
        request = isolate_proto.SetSecretRequest(name=name, value=None)
        self.stub.SetSecret(request)

    def list_secrets(self) -> list[ServerlessSecret]:
        request = isolate_proto.ListSecretsRequest()
        response = self.stub.ListSecrets(request)
        return [
            ServerlessSecret(
                name=secret.name,
                created_at=isolate_proto.datetime_from_timestamp(secret.created_time),
            )
            for secret in response.secrets
        ]



from typing import Any, Dict, List, Type, TypeVar

import attr

T = TypeVar("T", bound="GroupedUsageDetail")


@attr.s(auto_attribs=True)
class GroupedUsageDetail:
    """
    Attributes:
        model_id (str):
        machine_type (str):
        request_count (int):
        avg_duration (float):
        total_duration (float):
    """

    model_id: str
    machine_type: str
    request_count: int
    avg_duration: float
    total_duration: float
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        model_id = self.model_id
        machine_type = self.machine_type
        request_count = self.request_count
        avg_duration = self.avg_duration
        total_duration = self.total_duration

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "model_id": model_id,
                "machine_type": machine_type,
                "request_count": request_count,
                "avg_duration": avg_duration,
                "total_duration": total_duration,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        model_id = d.pop("model_id")

        machine_type = d.pop("machine_type")

        request_count = d.pop("request_count")

        avg_duration = d.pop("avg_duration")

        total_duration = d.pop("total_duration")

        grouped_usage_detail = cls(
            model_id=model_id,
            machine_type=machine_type,
            request_count=request_count,
            avg_duration=avg_duration,
            total_duration=total_duration,
        )

        grouped_usage_detail.additional_properties = d
        return grouped_usage_detail

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import time
from dataclasses import dataclass, field
from typing import Any, Iterator

import httpx
from fal import flags
from fal.sdk import Credentials, get_default_credentials

_URL_FORMAT = f"https://{{app_id}}.{flags.GATEWAY_HOST}/fal/queue"


@dataclass
class _Status:
    ...


@dataclass
class Queued(_Status):
    """Indicates the request is still in the queue, and provides the position
    in the queue for ETA calculation."""

    position: int


@dataclass
class InProgress(_Status):
    """Indicates the request is now being actively processed, and provides runtime
    logs for the inference task."""

    logs: list[dict[str, Any]] = field(default_factory=list)


@dataclass
class Completed(_Status):
    """Indicates the request has been completed successfully and the result is
    ready to be retrieved."""


@dataclass
class RequestHandle:
    """A handle to an async inference request."""

    app_id: str
    request_id: str

    # Use the credentials that were used to submit the request by default.
    _creds: Credentials = field(default_factory=get_default_credentials, repr=False)

    def status(self) -> _Status:
        """Check the status of an async inference request."""

        url = (
            _URL_FORMAT.format(app_id=self.app_id)
            + f"/requests/{self.request_id}/status/"
        )
        response = _HTTP_CLIENT.get(url, headers=self._creds.to_headers())
        response.raise_for_status()

        if response.status_code == 200:
            return Completed()

        data = response.json()
        if data["status"] == "IN_QUEUE":
            return Queued(position=data["queue_position"])
        elif data["status"] == "IN_PROGRESS":
            return InProgress(logs=data["logs"])
        else:
            raise ValueError(f"Unknown status: {data['status']}")

    def iter_events(
        self,
        *,
        __poll_delay: float = 0.2,
    ) -> Iterator[_Status]:
        """Yield all events regarding the given task till its completed."""

        while True:
            status = self.status()

            if isinstance(status, Completed):
                return

            yield status
            time.sleep(__poll_delay)

    def fetch_result(self) -> dict[str, Any]:
        """Retrieve the result of an async inference request, raises an exception
        if the request is not completed yet."""
        url = (
            _URL_FORMAT.format(app_id=self.app_id)
            + f"/requests/{self.request_id}/response/"
        )
        response = _HTTP_CLIENT.get(url, headers=self._creds.to_headers())
        response.raise_for_status()

        data = response.json()
        return data

    def get(self) -> dict[str, Any]:
        """Retrieve the result of an async inference request, polling the status
        of the request until it is completed."""

        for event in self.iter_events():
            continue

        return self.fetch_result()


_HTTP_CLIENT = httpx.Client(headers={"User-Agent": "Fal/Python"})


def run(app_id: str, arguments: dict[str, Any], *, path: str = "/") -> dict[str, Any]:
    """Run an inference task on a Fal app and return the result."""

    handle = submit(app_id, arguments, path=path)
    return handle.get()


def submit(app_id: str, arguments: dict[str, Any], *, path: str = "/") -> RequestHandle:
    """Submit an async inference task to the app. Returns a request handle
    which can be used to check the status of the request and retrieve the
    result."""

    url = _URL_FORMAT.format(app_id=app_id) + "/submit" + path
    creds = get_default_credentials()

    response = _HTTP_CLIENT.post(
        url,
        json=arguments,
        headers=creds.to_headers(),
    )
    response.raise_for_status()

    data = response.json()
    return RequestHandle(
        app_id=app_id,
        request_id=data["request_id"],
        _creds=creds,
    )



""" Contains all the data models used in inputs/outputs """

from .body_upload_file import BodyUploadFile
from .body_upload_local_file import BodyUploadLocalFile
from .customer_details import CustomerDetails
from .file_spec import FileSpec
from .gateway_usage_stats import GatewayUsageStats
from .grouped_usage_detail import GroupedUsageDetail
from .handle_stripe_webhook_response_handle_stripe_webhook import HandleStripeWebhookResponseHandleStripeWebhook
from .hash_check import HashCheck
from .http_validation_error import HTTPValidationError
from .invoice import Invoice
from .invoice_item import InvoiceItem
from .key_scope import KeyScope
from .log_entry import LogEntry
from .log_entry_labels import LogEntryLabels
from .new_user_key import NewUserKey
from .payment_method import PaymentMethod
from .persisted_usage_record import PersistedUsageRecord
from .persisted_usage_record_meta import PersistedUsageRecordMeta
from .run_type import RunType
from .status import Status
from .status_health import StatusHealth
from .uploaded_file_result import UploadedFileResult
from .url_file_upload import UrlFileUpload
from .usage_per_machine_type import UsagePerMachineType
from .usage_run_detail import UsageRunDetail
from .user_key_info import UserKeyInfo
from .validation_error import ValidationError

__all__ = (
    "BodyUploadFile",
    "BodyUploadLocalFile",
    "CustomerDetails",
    "FileSpec",
    "GatewayUsageStats",
    "GroupedUsageDetail",
    "HandleStripeWebhookResponseHandleStripeWebhook",
    "HashCheck",
    "HTTPValidationError",
    "Invoice",
    "InvoiceItem",
    "KeyScope",
    "LogEntry",
    "LogEntryLabels",
    "NewUserKey",
    "PaymentMethod",
    "PersistedUsageRecord",
    "PersistedUsageRecordMeta",
    "RunType",
    "Status",
    "StatusHealth",
    "UploadedFileResult",
    "UrlFileUpload",
    "UsagePerMachineType",
    "UsageRunDetail",
    "UserKeyInfo",
    "ValidationError",
)



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="PaymentMethod")


@attr.s(auto_attribs=True)
class PaymentMethod:
    """
    Attributes:
        id (str):
        type (str):
        last4 (str):
        brand (str):
        exp_month (Union[Unset, int]):
        exp_year (Union[Unset, int]):
    """

    id: str
    type: str
    last4: str
    brand: str
    exp_month: Union[Unset, int] = UNSET
    exp_year: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        type = self.type
        last4 = self.last4
        brand = self.brand
        exp_month = self.exp_month
        exp_year = self.exp_year

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "type": type,
                "last4": last4,
                "brand": brand,
            }
        )
        if exp_month is not UNSET:
            field_dict["exp_month"] = exp_month
        if exp_year is not UNSET:
            field_dict["exp_year"] = exp_year

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        type = d.pop("type")

        last4 = d.pop("last4")

        brand = d.pop("brand")

        exp_month = d.pop("exp_month", UNSET)

        exp_year = d.pop("exp_year", UNSET)

        payment_method = cls(
            id=id,
            type=type,
            last4=last4,
            brand=brand,
            exp_month=exp_month,
            exp_year=exp_year,
        )

        payment_method.additional_properties = d
        return payment_method

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

from ._base import FalServerlessException


class UnauthenticatedException(FalServerlessException):
    """Exception that indicates that"""

    def __init__(self) -> None:
        super().__init__(
            message="You must be authenticated.",
            hint="Login via `fal-serverless auth login`",
        )



from __future__ import annotations

from fal import apps

# TODO: DEPRECATED - use function instead
from fal.api import FalServerlessHost, LocalHost, cached
from fal.api import function
from fal.api import function as isolated
from fal.decorators import download_file, download_repo, download_weights
from fal.sdk import FalServerlessKeyCredentials
from fal.sync import sync_dir

local = LocalHost()
serverless = FalServerlessHost()

# DEPRECATED - use serverless instead
cloud = FalServerlessHost()

DBT_FAL_IMPORT_NOTICE = """
The dbt tool `fal` and `dbt-fal` adapter have been merged into a single tool.
Please import from the `fal.dbt` module instead.
Running `pip install dbt-fal` will install the new tool and the adapter alongside.
Then import from the `fal.dbt` module like

    from fal.dbt import {name}

"""


# Avoid printing on non-direct imports
def __getattr__(name: str):
    if name in (
        "NodeStatus",
        "FalDbt",
        "DbtModel",
        "DbtSource",
        "DbtTest",
        "DbtGenericTest",
        "DbtSingularTest",
        "Context",
        "CurrentModel",
    ):
        raise ImportError(DBT_FAL_IMPORT_NOTICE.format(name=name))

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


# NOTE: taken from dbt-core: https://github.com/dbt-labs/dbt-core/blob/ac539fd5cf325cfb5315339077d03399d575f570/core/dbt/adapters/__init__.py#L1-L7
# N.B.
# This will add to the package’s __path__ all subdirectories of directories on sys.path named after the package which effectively combines both modules into a single namespace (dbt.adapters)
# The matching statement is in plugins/postgres/dbt/adapters/__init__.py

from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.status_health import StatusHealth
from ..types import UNSET, Unset

T = TypeVar("T", bound="Status")


@attr.s(auto_attribs=True)
class Status:
    """
    Attributes:
        health (Union[Unset, StatusHealth]):  Default: StatusHealth.HEALTHY.
        queue_size (Union[Unset, int]):
        median_processing_time (Union[Unset, float]):
    """

    health: Union[Unset, StatusHealth] = StatusHealth.HEALTHY
    queue_size: Union[Unset, int] = 0
    median_processing_time: Union[Unset, float] = 0.0
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        health: Union[Unset, str] = UNSET
        if not isinstance(self.health, Unset):
            health = self.health.value

        queue_size = self.queue_size
        median_processing_time = self.median_processing_time

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if health is not UNSET:
            field_dict["health"] = health
        if queue_size is not UNSET:
            field_dict["queue_size"] = queue_size
        if median_processing_time is not UNSET:
            field_dict["median_processing_time"] = median_processing_time

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        _health = d.pop("health", UNSET)
        health: Union[Unset, StatusHealth]
        if isinstance(_health, Unset):
            health = UNSET
        else:
            health = StatusHealth(_health)

        queue_size = d.pop("queue_size", UNSET)

        median_processing_time = d.pop("median_processing_time", UNSET)

        status = cls(
            health=health,
            queue_size=queue_size,
            median_processing_time=median_processing_time,
        )

        status.additional_properties = d
        return status

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import io
from typing import TYPE_CHECKING, Literal, Optional, Union

from fal.toolkit.file.file import DEFAULT_REPOSITORY, File
from fal.toolkit.file.types import FileData, FileRepository, RepositoryId
from fal.toolkit.mainify import mainify
from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from PIL import Image as PILImage


ImageSizePreset = Literal[
    "square_hd",
    "square",
    "portrait_4_3",
    "portrait_16_9",
    "landscape_4_3",
    "landscape_16_9",
]


@mainify
class ImageSize(BaseModel):
    width: int = Field(
        default=512, description="The width of the generated image.", gt=0, le=4096
    )
    height: int = Field(
        default=512, description="The height of the generated image.", gt=0, le=4096
    )


IMAGE_SIZE_PRESETS: dict[ImageSizePreset, ImageSize] = {
    "square_hd": ImageSize(width=1024, height=1024),
    "square": ImageSize(width=512, height=512),
    "portrait_4_3": ImageSize(width=768, height=1024),
    "portrait_16_9": ImageSize(width=576, height=1024),
    "landscape_4_3": ImageSize(width=1024, height=768),
    "landscape_16_9": ImageSize(width=1024, height=576),
}

ImageSizeInput = Union[ImageSize, ImageSizePreset]


def get_image_size(source: ImageSizeInput) -> ImageSize:
    if isinstance(source, ImageSize):
        return source
    if isinstance(source, str) and source in IMAGE_SIZE_PRESETS:
        size = IMAGE_SIZE_PRESETS[source]
        return size
    raise TypeError(f"Invalid value for ImageSize: {source}")


get_image_size.__module__ = "__main__"

ImageFormat = Literal["png", "jpeg", "jpg", "webp", "gif"]


@mainify
class Image(File):
    """
    Represents an image file.
    """

    width: Optional[int] = Field(
        description="The width of the image in pixels.",
        examples=[1024],
    )
    height: Optional[int] = Field(
        description="The height of the image in pixels.",
        examples=[1024]
    )

    @classmethod
    def from_bytes(  # type: ignore[override]
        cls,
        data: bytes,
        format: ImageFormat,
        size: ImageSize | None = None,
        file_name: str | None = None,
        repository: FileRepository | RepositoryId = DEFAULT_REPOSITORY,
    ) -> Image:
        file_data = FileData(
            data=data, content_type=f"image/{format}", file_name=file_name
        )
        return cls(
            file_data=file_data,
            repository=repository,
            width=size.width if size else None,
            height=size.height if size else None,
        )

    @classmethod
    def from_pil(
        cls,
        pil_image: PILImage.Image,
        format: ImageFormat | None = None,
        file_name: str | None = None,
        repository: FileRepository | RepositoryId = DEFAULT_REPOSITORY,
    ) -> Image:
        size = ImageSize(width=pil_image.width, height=pil_image.height)
        if format is None:
            format = pil_image.format or "png"  # type: ignore[assignment]
            assert format  # for type checker

        with io.BytesIO() as f:
            pil_image.save(f, format=format)
            raw_image = f.getvalue()

        return cls.from_bytes(raw_image, format, size, file_name, repository)



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

import attr
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.log_entry_labels import LogEntryLabels


T = TypeVar("T", bound="LogEntry")


@attr.s(auto_attribs=True)
class LogEntry:
    """
    Attributes:
        timestamp (datetime.datetime):
        level (str):
        message (str):
        app (str):
        labels (Union[Unset, LogEntryLabels]):
    """

    timestamp: datetime.datetime
    level: str
    message: str
    app: str
    labels: Union[Unset, "LogEntryLabels"] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        timestamp = self.timestamp.isoformat()

        level = self.level
        message = self.message
        app = self.app
        labels: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.labels, Unset):
            labels = self.labels.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "timestamp": timestamp,
                "level": level,
                "message": message,
                "app": app,
            }
        )
        if labels is not UNSET:
            field_dict["labels"] = labels

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.log_entry_labels import LogEntryLabels

        d = src_dict.copy()
        timestamp = isoparse(d.pop("timestamp"))

        level = d.pop("level")

        message = d.pop("message")

        app = d.pop("app")

        _labels = d.pop("labels", UNSET)
        labels: Union[Unset, LogEntryLabels]
        if isinstance(_labels, Unset):
            labels = UNSET
        else:
            labels = LogEntryLabels.from_dict(_labels)

        log_entry = cls(
            timestamp=timestamp,
            level=level,
            message=message,
            app=app,
            labels=labels,
        )

        log_entry.additional_properties = d
        return log_entry

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Dict, Optional, Union, cast

import httpx

from ... import errors
from ...client import Client
from ...models.http_validation_error import HTTPValidationError
from ...types import UNSET, Response


def _get_kwargs(
    *,
    client: Client,
    user_id: str,
    lock: bool,
) -> Dict[str, Any]:
    url = "{}/admin/users/lock".format(client.base_url)

    headers: Dict[str, str] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    params: Dict[str, Any] = {}
    params["user_id"] = user_id

    params["lock"] = lock

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    return {
        "method": "post",
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
        "follow_redirects": client.follow_redirects,
        "params": params,
    }


def _parse_response(*, client: Client, response: httpx.Response) -> Optional[Union[HTTPValidationError, str]]:
    if response.status_code == HTTPStatus.OK:
        response_200 = cast(str, response.json())
        return response_200
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = HTTPValidationError.from_dict(response.json())

        return response_422
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(*, client: Client, response: httpx.Response) -> Response[Union[HTTPValidationError, str]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Client,
    user_id: str,
    lock: bool,
) -> Response[Union[HTTPValidationError, str]]:
    """Handle User Lock

     Lock a user with a given ID. Returns the locked user id.

    Args:
        user_id (str):
        lock (bool):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[HTTPValidationError, str]]
    """

    kwargs = _get_kwargs(
        client=client,
        user_id=user_id,
        lock=lock,
    )

    response = httpx.request(
        verify=client.verify_ssl,
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Client,
    user_id: str,
    lock: bool,
) -> Optional[Union[HTTPValidationError, str]]:
    """Handle User Lock

     Lock a user with a given ID. Returns the locked user id.

    Args:
        user_id (str):
        lock (bool):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[HTTPValidationError, str]
    """

    return sync_detailed(
        client=client,
        user_id=user_id,
        lock=lock,
    ).parsed


async def asyncio_detailed(
    *,
    client: Client,
    user_id: str,
    lock: bool,
) -> Response[Union[HTTPValidationError, str]]:
    """Handle User Lock

     Lock a user with a given ID. Returns the locked user id.

    Args:
        user_id (str):
        lock (bool):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[HTTPValidationError, str]]
    """

    kwargs = _get_kwargs(
        client=client,
        user_id=user_id,
        lock=lock,
    )

    async with httpx.AsyncClient(verify=client.verify_ssl) as _client:
        response = await _client.request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Client,
    user_id: str,
    lock: bool,
) -> Optional[Union[HTTPValidationError, str]]:
    """Handle User Lock

     Lock a user with a given ID. Returns the locked user id.

    Args:
        user_id (str):
        lock (bool):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[HTTPValidationError, str]
    """

    return (
        await asyncio_detailed(
            client=client,
            user_id=user_id,
            lock=lock,
        )
    ).parsed



from typing import Any, Dict, List, Type, TypeVar

import attr

T = TypeVar("T", bound="GatewayUsageStats")


@attr.s(auto_attribs=True)
class GatewayUsageStats:
    """
    Attributes:
        application_id (str):
        application_alias (str):
        request_count (int):
        success_count (int):
        error_count (int):
        total_billable_duration (int):
        p25_duration (float):
        p50_duration (float):
        p75_duration (float):
        p90_duration (float):
        price (float):
    """

    application_id: str
    application_alias: str
    request_count: int
    success_count: int
    error_count: int
    total_billable_duration: int
    p25_duration: float
    p50_duration: float
    p75_duration: float
    p90_duration: float
    price: float
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        application_id = self.application_id
        application_alias = self.application_alias
        request_count = self.request_count
        success_count = self.success_count
        error_count = self.error_count
        total_billable_duration = self.total_billable_duration
        p25_duration = self.p25_duration
        p50_duration = self.p50_duration
        p75_duration = self.p75_duration
        p90_duration = self.p90_duration
        price = self.price

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "application_id": application_id,
                "application_alias": application_alias,
                "request_count": request_count,
                "success_count": success_count,
                "error_count": error_count,
                "total_billable_duration": total_billable_duration,
                "p25_duration": p25_duration,
                "p50_duration": p50_duration,
                "p75_duration": p75_duration,
                "p90_duration": p90_duration,
                "price": price,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        application_id = d.pop("application_id")

        application_alias = d.pop("application_alias")

        request_count = d.pop("request_count")

        success_count = d.pop("success_count")

        error_count = d.pop("error_count")

        total_billable_duration = d.pop("total_billable_duration")

        p25_duration = d.pop("p25_duration")

        p50_duration = d.pop("p50_duration")

        p75_duration = d.pop("p75_duration")

        p90_duration = d.pop("p90_duration")

        price = d.pop("price")

        gateway_usage_stats = cls(
            application_id=application_id,
            application_alias=application_alias,
            request_count=request_count,
            success_count=success_count,
            error_count=error_count,
            total_billable_duration=total_billable_duration,
            p25_duration=p25_duration,
            p50_duration=p50_duration,
            p75_duration=p75_duration,
            p90_duration=p90_duration,
            price=price,
        )

        gateway_usage_stats.additional_properties = d
        return gateway_usage_stats

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import inspect
import sys
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from contextlib import suppress
from dataclasses import dataclass, field, replace
from functools import partial, wraps
from os import PathLike
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterator,
    Literal,
    TypeVar,
    cast,
    overload,
)

import dill
import dill.detect
import fal.flags as flags
import grpc
import isolate
import yaml
from fal._serialization import add_serialization_listeners_for, patch_dill
from fal.logging.isolate import IsolateLogPrinter
from fal.sdk import (
    FAL_SERVERLESS_DEFAULT_KEEP_ALIVE,
    Credentials,
    FalServerlessClient,
    FalServerlessConnection,
    HostedRunState,
    MachineRequirements,
    _get_agent_credentials,
    get_default_credentials,
)
from isolate.backends.common import active_python
from isolate.backends.settings import DEFAULT_SETTINGS
from isolate.connections import PythonIPC
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from typing_extensions import Concatenate, ParamSpec

ArgsT = ParamSpec("ArgsT")
ReturnT = TypeVar("ReturnT", covariant=True)

BasicConfig = Dict[str, Any]
_UNSET = object()


@dataclass
class FalServerlessError(Exception):
    message: str


@dataclass
class InternalFalServerlessError(Exception):
    message: str


@dataclass
class FalMissingDependencyError(FalServerlessError):
    ...


@dataclass
class Host(Generic[ArgsT, ReturnT]):
    """The physical environment where the isolated code
    is executed."""

    _SUPPORTED_KEYS: ClassVar[frozenset[str]] = frozenset()
    _GATEWAY_KEYS: ClassVar[frozenset[str]] = frozenset(
        {"serve", "exposed_port", "max_concurrency"}
    )

    def __post_init__(self):
        assert not self._SUPPORTED_KEYS.intersection(
            self._GATEWAY_KEYS
        ), "Gateway keys and host keys should not overlap"

    @classmethod
    def parse_key(cls, key: str, value: Any) -> tuple[Any, Any]:
        if key == "env_yml":
            # Conda environment definition should be parsed before sending to serverless
            with open(value) as f:
                return "env_dict", yaml.safe_load(f)
        else:
            return key, value

    @classmethod
    def parse_options(cls, **config: Any) -> Options:
        """Split the given set of options into host and
        environment options."""

        options = Options()
        for key, value in config.items():
            key, value = cls.parse_key(key, value)
            if key in cls._SUPPORTED_KEYS:
                options.host[key] = value
            elif key in cls._GATEWAY_KEYS:
                options.gateway[key] = value
            else:
                options.environment[key] = value

        if options.gateway.get("serve"):
            options.add_requirements(["fastapi==0.99.1", "uvicorn"])

        return options

    def run(
        self,
        func: Callable[ArgsT, ReturnT],
        options: Options,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> ReturnT:
        """Run the given function in the isolated environment."""
        raise NotImplementedError


def cached(func: Callable[ArgsT, ReturnT]) -> Callable[ArgsT, ReturnT]:
    """Cache the result of the given function in-memory."""
    import hashlib

    try:
        source_code = inspect.getsource(func).encode("utf-8")
    except OSError:
        # TODO: explain the reason for this (e.g. we don't know how to
        # check if you sent us the same function twice).
        print(f"[warning] Function {func.__name__} can not be cached...")
        return func

    cache_key = hashlib.sha256(source_code).hexdigest()

    @wraps(func)
    def wrapper(
        *args: ArgsT.args,
        **kwargs: ArgsT.kwargs,
    ) -> ReturnT:
        from functools import lru_cache

        # HACK: Using the isolate module as a global cache.
        import isolate

        if not hasattr(isolate, "__cached_functions__"):
            isolate.__cached_functions__ = {}

        if cache_key not in isolate.__cached_functions__:
            isolate.__cached_functions__[cache_key] = lru_cache(maxsize=None)(func)

        return isolate.__cached_functions__[cache_key](*args, **kwargs)

    return wrapper


def _prepare_partial_func(
    func: Callable[ArgsT, ReturnT],
    *args: ArgsT.args,
    **kwargs: ArgsT.kwargs,
) -> Callable[ArgsT, ReturnT]:
    """Prepare the given function for execution on the remote isolate workers."""

    @wraps(func)
    def wrapper(*remote_args: ArgsT.args, **remote_kwargs: ArgsT.kwargs) -> ReturnT:
        result = func(*remote_args, *args, **remote_kwargs, **kwargs)
        with suppress(Exception):
            patch_dill()
        return result

    return wrapper


@dataclass
class LocalHost(Host):
    # The environment which provides the default set of
    # packages for isolate agent to run.
    _AGENT_ENVIRONMENT = isolate.prepare_environment(
        "virtualenv",
        requirements=[f"dill=={dill.__version__}"],
    )

    def run(
        self,
        func: Callable[..., ReturnT],
        options: Options,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> ReturnT:
        settings = replace(DEFAULT_SETTINGS, serialization_method="dill")
        environment = isolate.prepare_environment(
            **options.environment,
            context=settings,
        )
        with PythonIPC(
            environment,
            environment.create(),
            extra_inheritance_paths=[self._AGENT_ENVIRONMENT.create()],
        ) as connection:
            executable = partial(func, *args, **kwargs)
            return connection.run(executable)


FAL_SERVERLESS_DEFAULT_URL = flags.GRPC_HOST
FAL_SERVERLESS_DEFAULT_MACHINE_TYPE = "XS"


import threading


def _handle_grpc_error():
    def decorator(fn):
        @wraps(fn)
        def handler(*args, **kwargs):
            """
            Wraps grpc errors as fal Serverless Errors.
            """
            try:
                return fn(*args, **kwargs)
            except grpc.RpcError as e:
                if e.code() == grpc.StatusCode.UNAVAILABLE:
                    raise FalServerlessError(
                        "Could not reach fal Serverless host. "
                        "This is most likely a transient problem. "
                        "Please, try again."
                    )
                elif e.details().endswith("died with <Signals.SIGKILL: 9>.`."):
                    raise FalServerlessError(
                        "Isolated function crashed. "
                        "This is likely due to resource overflow. "
                        "You can try again by setting a bigger `machine_type`"
                    )

                elif e.code() == grpc.StatusCode.INVALID_ARGUMENT and (
                    "The function function could not be deserialized" in e.details()
                ):
                    raise FalMissingDependencyError(e.details()) from None
                else:
                    raise FalServerlessError(e.details())

        return handler

    return decorator


def find_missing_dependencies(
    func: Callable, env: dict
) -> Iterator[tuple[str, list[str]]]:
    if env["kind"] != "virtualenv":
        return

    used_modules = defaultdict(list)
    scope = {**dill.detect.globalvars(func, recurse=True), **dill.detect.freevars(func)}  # type: ignore

    for name, obj in scope.items():
        if isinstance(obj, IsolatedFunction):
            used_modules["fal"].append(name)
            continue

        module = inspect.getmodule(obj)
        if module is None:
            continue

        possible_package = getattr(module, "__package__", None)
        if possible_package:
            pkg_name, *_ = possible_package.split(".")  # type: ignore
        else:
            pkg_name = module.__name__

        used_modules[canonicalize_name(pkg_name)].append(name)  # type: ignore

    raw_requirements = env.get("requirements", [])
    specified_requirements = set()
    for raw_requirement in raw_requirements:
        try:
            requirement = Requirement(raw_requirement)
        except ValueError:
            # For git+ dependencies, we can't parse the canonical name
            # so we'll just skip them.
            continue
        specified_requirements.add(canonicalize_name(requirement.name))

    for module_name, used_names in used_modules.items():
        if module_name in specified_requirements:
            continue
        yield module_name, used_names


# TODO: Should we build all these in fal/dbt-fal packages instead?
@dataclass
class FalServerlessHost(Host):
    _SUPPORTED_KEYS = frozenset(
        {
            "machine_type",
            "keep_alive",
            "setup_function",
            "_base_image",
            "_scheduler",
        }
    )

    url: str = FAL_SERVERLESS_DEFAULT_URL
    credentials: Credentials = field(default_factory=get_default_credentials)
    _lock: threading.Lock = field(default_factory=threading.Lock)

    _log_printer = IsolateLogPrinter(debug=flags.DEBUG)

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.__dict__.update(state)
        self.credentials = _get_agent_credentials(self.credentials)

    @property
    def _connection(self) -> FalServerlessConnection:
        with self._lock:
            client = FalServerlessClient(self.url, self.credentials)
            return client.connect()

    @_handle_grpc_error()
    def register(
        self,
        func: Callable[ArgsT, ReturnT],
        options: Options,
        max_concurrency: int | None = None,
        application_name: str | None = None,
        application_auth_mode: Literal["public", "shared", "private"] | None = None,
    ) -> str | None:
        environment_options = options.environment.copy()
        environment_options.setdefault("python_version", active_python())
        environments = [self._connection.define_environment(**environment_options)]

        machine_type = options.host.get(
            "machine_type", FAL_SERVERLESS_DEFAULT_MACHINE_TYPE
        )
        keep_alive = options.host.get("keep_alive", FAL_SERVERLESS_DEFAULT_KEEP_ALIVE)
        base_image = options.host.get("_base_image", None)
        scheduler = options.host.get("_scheduler", None)
        exposed_port = options.get_exposed_port()

        machine_requirements = MachineRequirements(
            machine_type=machine_type,
            keep_alive=keep_alive,
            base_image=base_image,
            exposed_port=exposed_port,
            scheduler=scheduler,
        )

        partial_func = _prepare_partial_func(func)

        for partial_result in self._connection.register(
            partial_func,
            environments,
            application_name=application_name,
            application_auth_mode=application_auth_mode,
            machine_requirements=machine_requirements,
            max_concurrency=max_concurrency,
        ):
            for log in partial_result.logs:
                self._log_printer.print(log)

            if partial_result.result:
                return partial_result.result.application_id

    @_handle_grpc_error()
    def schedule(
        self,
        func: Callable[ArgsT, ReturnT],
        cron: str,
        options: Options,
    ) -> str | None:
        application_id = self.register(func, options)
        if application_id is None:
            return None
        return self._connection.schedule_cronjob(application_id, cron)

    @_handle_grpc_error()
    def run(
        self,
        func: Callable[..., ReturnT],
        options: Options,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> ReturnT:
        environment_options = options.environment.copy()
        environment_options.setdefault("python_version", active_python())
        environments = [self._connection.define_environment(**environment_options)]

        machine_type = options.host.get(
            "machine_type", FAL_SERVERLESS_DEFAULT_MACHINE_TYPE
        )
        keep_alive = options.host.get("keep_alive", FAL_SERVERLESS_DEFAULT_KEEP_ALIVE)
        base_image = options.host.get("_base_image", None)
        scheduler = options.host.get("_scheduler", None)
        exposed_port = options.get_exposed_port()
        setup_function = options.host.get("setup_function", None)

        machine_requirements = MachineRequirements(
            machine_type=machine_type,
            keep_alive=keep_alive,
            base_image=base_image,
            exposed_port=exposed_port,
            scheduler=scheduler,
        )

        return_value = _UNSET
        # Allow isolate provided arguments (such as setup function) to take
        # precedence over the ones provided by the user.
        partial_func = _prepare_partial_func(func, *args, **kwargs)
        for partial_result in self._connection.run(
            partial_func,
            environments,
            machine_requirements=machine_requirements,
            setup_function=setup_function,
        ):
            for log in partial_result.logs:
                self._log_printer.print(log)

            if partial_result.status.state is not HostedRunState.IN_PROGRESS:
                state = partial_result.status.state
                if state is HostedRunState.INTERNAL_FAILURE:
                    raise InternalFalServerlessError(
                        "An internal failure occurred while performing this run."
                    )
                elif state is HostedRunState.SUCCESS:
                    return_value = partial_result.result
                else:
                    raise NotImplementedError("Unknown state: ", state)

        if return_value is _UNSET:
            raise InternalFalServerlessError(
                "The input function did not return any value."
            )

        return cast(ReturnT, return_value)


@dataclass
class Options:
    host: BasicConfig = field(default_factory=dict)
    environment: BasicConfig = field(default_factory=dict)
    gateway: BasicConfig = field(default_factory=dict)

    def add_requirements(self, requirements: list[str]):
        kind = self.environment["kind"]
        if kind == "virtualenv":
            pip_requirements = self.environment.setdefault("requirements", [])
        elif kind == "conda":
            pip_requirements = self.environment.setdefault("pip", [])
        else:
            raise FalServerlessError(
                "Only conda and virtualenv is supported as environment options"
            )

        # Already has these.
        if set(pip_requirements).issuperset(set(requirements)):
            return None

        pip_requirements.extend(requirements)

    def get_exposed_port(self) -> int | None:
        if self.gateway.get("serve"):
            return _SERVE_PORT
        else:
            return self.gateway.get("exposed_port")


_DEFAULT_HOST = FalServerlessHost()
_SERVE_PORT = 8080

# Overload @function to help users identify the correct signature.
# NOTE: This is both in sync with host options and with environment configs from `isolate` package.


## virtualenv
### LocalHost
@overload
def function(
    kind: Literal["virtualenv"] = "virtualenv",
    *,
    python_version: str | None = None,
    requirements: list[str] | None = None,
    # Common options
    host: LocalHost,
    serve: Literal[False] = False,
    exposed_port: int | None = None,
    max_concurrency: int | None = None,
) -> Callable[
    [Callable[Concatenate[ArgsT], ReturnT]], IsolatedFunction[ArgsT, ReturnT]
]:
    ...


@overload
def function(
    kind: Literal["virtualenv"] = "virtualenv",
    *,
    python_version: str | None = None,
    requirements: list[str] | None = None,
    # Common options
    host: LocalHost,
    serve: Literal[True],
    exposed_port: int | None = None,
    max_concurrency: int | None = None,
) -> Callable[
    [Callable[Concatenate[ArgsT], ReturnT]], ServedIsolatedFunction[ArgsT, ReturnT]
]:
    ...


### FalServerlessHost
@overload
def function(
    kind: Literal["virtualenv"] = "virtualenv",
    *,
    python_version: str | None = None,
    requirements: list[str] | None = None,
    # Common options
    host: FalServerlessHost = _DEFAULT_HOST,
    serve: Literal[False] = False,
    exposed_port: int | None = None,
    max_concurrency: int | None = None,
    # FalServerlessHost options
    machine_type: str = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
    keep_alive: int = FAL_SERVERLESS_DEFAULT_KEEP_ALIVE,
    setup_function: Callable[..., None] | None = None,
    _base_image: str | None = None,
    _scheduler: str | None = None,
) -> Callable[
    [Callable[Concatenate[ArgsT], ReturnT]], IsolatedFunction[ArgsT, ReturnT]
]:
    ...


@overload
def function(
    kind: Literal["virtualenv"] = "virtualenv",
    *,
    python_version: str | None = None,
    requirements: list[str] | None = None,
    # Common options
    host: FalServerlessHost = _DEFAULT_HOST,
    serve: Literal[True],
    exposed_port: int | None = None,
    max_concurrency: int | None = None,
    # FalServerlessHost options
    machine_type: str = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
    keep_alive: int = FAL_SERVERLESS_DEFAULT_KEEP_ALIVE,
    setup_function: Callable[..., None] | None = None,
    _base_image: str | None = None,
    _scheduler: str | None = None,
) -> Callable[
    [Callable[Concatenate[ArgsT], ReturnT]], ServedIsolatedFunction[ArgsT, ReturnT]
]:
    ...


## conda
### LocalHost
@overload
def function(
    kind: Literal["conda"],
    *,
    python_version: str | None = None,
    env_dict: dict[str, Any] | None = None,
    env_yml: PathLike | str | None = None,
    env_yml_str: str | None = None,
    packages: list[str] | None = None,
    pip: list[str] | None = None,
    channels: list[str] | None = None,
    # Common options
    host: LocalHost,
    serve: Literal[False] = False,
    exposed_port: int | None = None,
    max_concurrency: int | None = None,
) -> Callable[
    [Callable[Concatenate[ArgsT], ReturnT]], IsolatedFunction[ArgsT, ReturnT]
]:
    ...


@overload
def function(
    kind: Literal["conda"],
    *,
    python_version: str | None = None,
    env_dict: dict[str, Any] | None = None,
    env_yml: PathLike | str | None = None,
    env_yml_str: str | None = None,
    packages: list[str] | None = None,
    pip: list[str] | None = None,
    channels: list[str] | None = None,
    # Common options
    host: LocalHost,
    serve: Literal[True],
    exposed_port: int | None = None,
    max_concurrency: int | None = None,
) -> Callable[
    [Callable[Concatenate[ArgsT], ReturnT]], ServedIsolatedFunction[ArgsT, ReturnT]
]:
    ...


### FalServerlessHost
@overload
def function(
    kind: Literal["conda"],
    *,
    python_version: str | None = None,
    env_dict: dict[str, Any] | None = None,
    env_yml: PathLike | str | None = None,
    env_yml_str: str | None = None,
    packages: list[str] | None = None,
    pip: list[str] | None = None,
    channels: list[str] | None = None,
    # Common options
    host: FalServerlessHost = _DEFAULT_HOST,
    serve: Literal[False] = False,
    exposed_port: int | None = None,
    max_concurrency: int | None = None,
    # FalServerlessHost options
    machine_type: str = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
    keep_alive: int = FAL_SERVERLESS_DEFAULT_KEEP_ALIVE,
    setup_function: Callable[..., None] | None = None,
    _base_image: str | None = None,
    _scheduler: str | None = None,
) -> Callable[
    [Callable[Concatenate[ArgsT], ReturnT]], IsolatedFunction[ArgsT, ReturnT]
]:
    ...


@overload
def function(
    kind: Literal["conda"],
    *,
    python_version: str | None = None,
    env_dict: dict[str, Any] | None = None,
    env_yml: PathLike | str | None = None,
    env_yml_str: str | None = None,
    packages: list[str] | None = None,
    pip: list[str] | None = None,
    channels: list[str] | None = None,
    # Common options
    host: FalServerlessHost = _DEFAULT_HOST,
    serve: Literal[True],
    exposed_port: int | None = None,
    max_concurrency: int | None = None,
    # FalServerlessHost options
    machine_type: str = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
    keep_alive: int = FAL_SERVERLESS_DEFAULT_KEEP_ALIVE,
    setup_function: Callable[..., None] | None = None,
    _base_image: str | None = None,
    _scheduler: str | None = None,
) -> Callable[
    [Callable[Concatenate[ArgsT], ReturnT]], ServedIsolatedFunction[ArgsT, ReturnT]
]:
    ...


# implementation
def function(  # type: ignore
    kind: str = "virtualenv",
    *,
    host: Host = _DEFAULT_HOST,
    **config: Any,
):
    options = host.parse_options(kind=kind, **config)

    def wrapper(func: Callable[ArgsT, ReturnT]):
        add_serialization_listeners_for(func)
        proxy = IsolatedFunction(
            host=host,
            raw_func=func,  # type: ignore
            options=options,
        )
        return wraps(func)(proxy)  # type: ignore

    return wrapper


def create_webapp(func: Callable[ArgsT, ReturnT]):
    def wrapper():
        from fastapi import FastAPI
        from fastapi.middleware.cors import CORSMiddleware
        from uvicorn import run

        app = FastAPI()

        app.add_middleware(
            CORSMiddleware,
            allow_credentials=True,
            allow_headers=("*"),
            allow_methods=("*"),
            allow_origins=("*"),
        )

        app.add_api_route("/", func, name=func.__name__, methods=["POST"])

        run(app, host="0.0.0.0", port=8080)

    return wrapper


@dataclass
class IsolatedFunction(Generic[ArgsT, ReturnT]):
    host: Host[ArgsT, ReturnT]
    raw_func: Callable[ArgsT, ReturnT]
    options: Options
    executor: ThreadPoolExecutor = field(default_factory=ThreadPoolExecutor)

    def __getstate__(self) -> dict[str, Any]:
        # Ensure that the executor is not pickled.
        state = self.__dict__.copy()
        del state["executor"]
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.__dict__.update(state)
        if not hasattr(self, "executor"):
            self.executor = ThreadPoolExecutor()

    def submit(self, *args: ArgsT.args, **kwargs: ArgsT.kwargs):
        # TODO: This should probably live inside each host since they can
        # have more optimized Future implementations (e.g. instead of real
        # threads, they can use state observers and detached runs).

        future = self.executor.submit(
            self.host.run,
            func=self.func,
            options=self.options,
            args=args,
            kwargs=kwargs,
        )
        return future

    def __call__(self, *args: ArgsT.args, **kwargs: ArgsT.kwargs) -> ReturnT:
        try:
            return self.host.run(
                self.func,
                self.options,
                args=args,
                kwargs=kwargs,
            )
        except FalMissingDependencyError as e:
            pairs = list(find_missing_dependencies(self.func, self.options.environment))
            if not pairs:
                raise e
            else:
                lines = []
                for used_modules, references in pairs:
                    lines.append(
                        f"\t- {used_modules!r} (accessed through {', '.join(map(repr, references))})"
                    )

                function_name = self.func.__name__
                raise FalServerlessError(
                    f"Couldn't deserialize your function on the remote server. \n\n[Hint] {function_name!r} "
                    f"function uses the following modules which weren't present in the environment definition:\n"
                    + "\n".join(lines)
                ) from None

    @overload
    def on(
        self, host: Host | None = None, *, serve: Literal[False] = False, **config: Any
    ) -> IsolatedFunction[ArgsT, ReturnT]:
        ...

    @overload
    def on(
        self, host: Host | None = None, *, serve: Literal[True], **config: Any
    ) -> ServedIsolatedFunction[ArgsT, ReturnT]:
        ...

    def on(self, host: Host | None = None, **config: Any):  # type: ignore
        host = host or self.host
        previous_host_options = {
            # Only keep the options that are supported by the new host.
            key: self.options.host.get(key)
            for key in host._SUPPORTED_KEYS
            if self.options.host.get(key) is not None
        }

        # The order of the options is important here (the latter
        # options override the former ones).
        host_options = {
            # All the previous options
            **previous_host_options,
            **self.options.environment,
            **self.options.gateway,
            # The new options
            **config,
        }
        new_options = host.parse_options(**host_options)
        return replace(
            self,
            host=host,
            options=new_options,
        )

    @property
    def func(self) -> Callable[ArgsT, ReturnT]:
        serve_mode = self.options.gateway.get("serve")
        if serve_mode:
            # This type can be safely ignored because this case only happens when it is a ServedIsolatedFunction, which correctly
            # has ArgsT = [] and ReturnT = None
            return create_webapp(self.raw_func)  # type: ignore
        else:
            return self.raw_func


if sys.version_info <= (3, 10):
    compatible_class = IsolatedFunction[Literal[None], None]  # type: ignore
else:
    compatible_class = IsolatedFunction[[], None]


class ServedIsolatedFunction(
    Generic[ArgsT, ReturnT],
    compatible_class,  # type: ignore
):
    # Class for type hinting purposes only.
    @overload  # type: ignore[override,no-overload-impl]
    def on(  # type: ignore[no-overload-impl]
        self, host: Host | None = None, *, serve: Literal[True] = True, **config: Any
    ) -> ServedIsolatedFunction[ArgsT, ReturnT]:
        ...

    @overload
    def on(
        self, host: Host | None = None, *, serve: Literal[False], **config: Any
    ) -> IsolatedFunction[ArgsT, ReturnT]:
        ...
